/*
 * Error diffusion over constrained bitmap
 * It is specially designed for "woven fabric display".
 * 2017-02-05: coded by Abdalla G. M. Ahmed
 * This is a grayscale only version
 * 2017-02-09: I abandoned the tone gradient idea that gave inferior results
 *  (more DC), and replaced it by the 8-matrices approach.
 */

#include "weaves-4.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string>
#include <getopt.h>
#include <vector>

int w, h;
unsigned char warp[4] = {1, 1, 1, 1};
unsigned char weft[4] = {0, 0, 0, 0};

double palette[] = {0.0, 1.0};
std::vector<double> img;
std::vector<int> bmp;


const double zero = 0.25;
const double rang = 0.5;

void loadImage(const char *fileName) {
    FILE *file = fopen(fileName, "r");
    int max;
    fscanf(file, "P2 %d %d %d ", &w, &h, &max);
    img.resize(w*h);
    int clr;
    for (int i = 0; i < h*w; i++) {
        fscanf(file, " %d", &clr);
        if (feof(file)) {
            fprintf(stderr, "Failed to load image\n");
            exit(1);
        }
        img[i] = rang * (double(max - clr) / max) + zero;                       // Tone has opposite senses in PGM and PBM files.
    }
    fclose(file);
}

void dither(unsigned int matrixNumber) {
    struct TDividend { int dx, dy; double share; };                             // For error diffusion
    const TDividend LTR[] = {
        {1, 0, 7./16.}, {-1, 1, 1./16.}, {0, 1, 5./16}, {1, 1, 3./16}
    };
    const TDividend RTL[] = {
        {-1, 0, 7./16.}, {1, 1, 1./16.}, {0, 1, 5./16}, {-1, 1, 3./16}
    };
    const int nShares = 4;
    const unsigned char *matrix = WEAVE[matrixNumber];
    unsigned char opt[2];
    bmp.resize(h * w);
    for (int y = 0; y < h; y++) {
        int start, stop, dx;
        const TDividend *dividends;
        if ((y&1) == 0) {
            start = 0; stop = w; dx = 1; dividends = LTR;
        }
        else {
            start = w-1; stop = -1; dx = -1; dividends = RTL;
        }
        for (int x = start; x != stop; x += dx) {
            int index = y * w + x;
            double &clr = img[index];
            int nOptions = 2;
            int options = matrix[4 * (y&3) + (x&3)];
            switch (options) {
                case 0: nOptions = 1; opt[0] = warp[x&3]; break;
                case 1: nOptions = 1; opt[0] = weft[y&3]; break;
                default: opt[0] = warp[x&3]; opt[1] = weft[y&3];
            }
            unsigned char best = opt[0];
            double error = clr - palette[best];
            for (int i = 1; i < nOptions; i++) {
                double tmp = clr - palette[ opt[i] ];
                if (fabs(tmp) < fabs(error)) {
                    best = opt[i];
                    error = tmp;
                }
            }
            bmp[index] = best;
            // Diffuse error:
            for (int i = 0; i < nShares; i++) {
                int nx = x + dividends[i].dx, ny = y + dividends[i].dy;         // Neighbor pixel coordinates
                if (nx >= 0 && nx < w && ny >= 0 && ny < h) {
                    img[ny * w + nx] += dividends[i].share * error;
                }
            }
        }
    }
}


void saveImage(const char *fileName) {
    FILE *file = fopen(fileName, "w");
    fprintf(file, "P1\n%d %d\n", w, h);
    for (int i = 0; i < h * w; i++) {
        fprintf(file, "%c\n", bmp[i] + '0');
    }
    fclose(file);
}

const char *USAGE_MESSAGE = "Usage: %s <pgm-file>\n"
    "The pgm file must not contain comments\n"
    "Options\n:"
;

int main(int argc, char **argv) {
    int opt;                                                                    // For use by getopt, the command line options utility
    int matrixNumber = 2;
    while ((opt = getopt(argc, argv, "m:")) != -1) {                             // Modify default settings with command line options
        switch (opt) {
            case 'm': matrixNumber = atoi(optarg); break;
            default: fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
        }
    }
    if (optind > argc - 1) {
        fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
    }
    std::string fileName = argv[optind];
    loadImage(fileName.c_str());
    dither(matrixNumber);
    fileName.replace(fileName.length() - 3, 3, "pbm");
    saveImage(fileName.c_str());
}
